@@ -24,6 +24,7 @@
 #include <algorithm>
 
 extern "C" {
+#include <unistd.h>
 #include <linux/limits.h>
 #include <sys/time.h>
 #include <glob.h>
@@ -133,6 +134,7 @@ struct compile_server_info
 };
 
 ostream &operator<< (ostream &s, const compile_server_info &i);
+ostream &operator<< (ostream &s, const vector<compile_server_info> &v);
 
 static void
 preferred_order (vector<compile_server_info> &servers)
@@ -752,7 +754,7 @@ compile_server_client::passes_0_4 ()
   PROBE1(stap, client__start, &s);
 
   // arguments parsed; get down to business
-  if (s.verbose > 1)
+  if (s.verbose)
     clog << _("Using a compile server.") << endl;
 
   struct tms tms_before;
@@ -824,13 +826,13 @@ compile_server_client::passes_0_4 ()
 	{
 	  string module_src_path = s.tmpdir + "/" + s.module_name + ".ko";
 	  string module_dest_path = s.module_name + ".ko";
-	  copy_file (module_src_path, module_dest_path, s.verbose > 1);
+	  copy_file (module_src_path, module_dest_path, s.verbose >= 3);
 	  // Also copy the module signature, it it exists.
 	  module_src_path += ".sgn";
 	  if (file_exists (module_src_path))
 	    {
 	      module_dest_path += ".sgn";
-	      copy_file(module_src_path, module_dest_path, s.verbose > 1);
+	      copy_file(module_src_path, module_dest_path, s.verbose >= 3);
 	    }
 	}
     }
@@ -1150,7 +1152,7 @@ compile_server_client::find_and_connect_to_server ()
 	  if (i->port == 0)
 	    {
 	      get_or_keep_compatible_server_info (s, online_servers, true/*keep*/);
-	      if (s.unprivileged)
+	      if (! pr_contains (s.privilege, pr_stapdev))
 		get_or_keep_signing_server_info (s, online_servers, true/*keep*/);
 	    }
 
@@ -1166,7 +1168,31 @@ compile_server_client::find_and_connect_to_server ()
   unsigned limit = server_list.size ();
   if (limit == 0)
     {
-      clog << _("Unable to find a compile server.") << endl;
+      clog << _("Unable to find a suitable compile server.") << endl;
+
+      // Try to explain why.
+      vector<compile_server_info> online_servers;
+      get_or_keep_online_server_info (s, online_servers, false/*keep*/);
+      if (online_servers.empty ())
+	clog << _("No servers online to select from.") << endl;
+      else
+	{
+	  clog << _("The following servers are online:") << endl;
+	  clog << online_servers;
+	  if (! specified_servers.empty ())
+	    {
+	      clog << _("The following servers were requested:") << endl;
+	      clog << specified_servers;
+	    }
+	  else
+	    {
+	      string criteria = "online,trusted,compatible";
+	      if (! pr_contains (s.privilege, pr_stapdev))
+		criteria += ",signer";
+	      clog << _F("No servers matched the selection criteria of %s.", criteria.c_str())
+		   << endl;
+	    }
+	}
       return 1;
     }
 
@@ -1183,8 +1209,8 @@ compile_server_client::find_and_connect_to_server ()
   // Give the server a chance to do this before retrying.
   if (rc == SERVER_CERT_EXPIRED_ERROR)
     {
-      if (s.verbose > 1)
-	clog << _("A server's certificate was expired. Trying again") << endl << flush;
+      if (s.verbose >= 2)
+	clog << _("The server's certificate was expired. Trying again") << endl << flush;
       sleep (2);
       rc = compile_using_server (server_list);
       if (rc == SUCCESS)
@@ -1193,6 +1219,8 @@ compile_server_client::find_and_connect_to_server ()
 
   // We were unable to use any available server
   clog << _("Unable to connect to a server.") << endl;
+  clog << _("The following servers were tried:") << endl;
+  clog << server_list;
   return 1; // Failure
 }
 
@@ -1252,7 +1280,7 @@ compile_server_client::compile_using_server (
 	   j != servers.end ();
 	   ++j)
 	{
-	  if (s.verbose > 1)
+	  if (s.verbose >= 2)
            clog << _F("Attempting SSL connection with %s\n"
                 "  using certificates from the database in %s\n",
                 lex_cast(*j).c_str(), cert_dir);
@@ -1290,7 +1318,7 @@ compile_server_client::compile_using_server (
 	      continue;
 	    }
 
-	  if (s.verbose > 1)
+	  if (s.verbose >= 2)
 	    {
 	      clog << _("  Unable to connect: ");
 	      nssError ();
@@ -1344,7 +1372,7 @@ compile_server_client::unpack_response ()
   // its contents to our temp directory.
   glob_t globbuf;
   string filespec = server_tmpdir + "/stap??????";
-  if (s.verbose > 2)
+  if (s.verbose >= 3)
     clog << _F("Searching \"%s\"\n", filespec.c_str());
   int r = glob(filespec.c_str (), 0, NULL, & globbuf);
   if (r != GLOB_NOSPACE && r != GLOB_ABORTED && r != GLOB_NOMATCH)
@@ -1358,11 +1386,11 @@ compile_server_client::unpack_response ()
 
       assert (globbuf.gl_pathc == 1);
       string dirname = globbuf.gl_pathv[0];
-      if (s.verbose > 2)
+      if (s.verbose >= 3)
 	clog << _("  found ") << dirname << endl;
 
       filespec = dirname + "/*";
-      if (s.verbose > 2)
+      if (s.verbose >= 3)
        clog << _F("Searching \"%s\"\n", filespec.c_str());
       int r = glob(filespec.c_str (), GLOB_PERIOD, NULL, & globbuf);
       if (r != GLOB_NOSPACE && r != GLOB_ABORTED && r != GLOB_NOMATCH)
@@ -1375,7 +1403,7 @@ compile_server_client::unpack_response ()
 		  oldname.substr (oldname.size () - 3) == "/..")
 		continue;
 	      string newname = s.tmpdir + "/" + oldname.substr (prefix_len);
-	      if (s.verbose > 2)
+	      if (s.verbose >= 3)
                clog << _F("  found %s -- linking from %s", oldname.c_str(), newname.c_str());
 	      rc = symlink (oldname.c_str (), newname.c_str ());
 	      if (rc != 0)
@@ -1430,7 +1458,7 @@ compile_server_client::process_response ()
     {
       // The server should have returned a module.
       string filespec = s.tmpdir + "/*.ko";
-      if (s.verbose > 2)
+      if (s.verbose >= 3)
        clog << _F("Searching \"%s\"\n", filespec.c_str());
 
       glob_t globbuf;
@@ -1443,7 +1471,7 @@ compile_server_client::process_response ()
 	    {
 	      assert (globbuf.gl_pathc == 1);
 	      string modname = globbuf.gl_pathv[0];
-	      if (s.verbose > 2)
+	      if (s.verbose >= 3)
 		clog << _("  found ") << modname << endl;
 
 	      // If a module name was not specified by the user, then set it to
@@ -1709,7 +1737,7 @@ add_server_trust (
       if (find (already_trusted.begin (), already_trusted.end (), *server) !=
 	  already_trusted.end ())
 	{
-	  if (s.verbose > 1)
+	  if (s.verbose >= 2)
 	    trust_already_in_place (*server, server_list, cert_db_path, false/*revoking*/);
 	  continue;
 	}
@@ -1737,7 +1765,7 @@ add_server_trust (
   // Make sure the database files are readable.
   glob_t globbuf;
   string filespec = cert_db_path + "/*.db";
-  if (s.verbose > 2)
+  if (s.verbose >= 3)
     clog << _F("Searching \"%s\"\n", filespec.c_str());
   int r = glob (filespec.c_str (), 0, NULL, & globbuf);
   if (r != GLOB_NOSPACE && r != GLOB_ABORTED && r != GLOB_NOMATCH)
@@ -1745,12 +1773,12 @@ add_server_trust (
       for (unsigned i = 0; i < globbuf.gl_pathc; ++i)
 	{
 	  string filename = globbuf.gl_pathv[i];
-	  if (s.verbose > 2)
+	  if (s.verbose >= 3)
 	    clog << _("  found ") << filename << endl;
 
 	  if (chmod (filename.c_str (), 0644) != 0)
 	    {
-             clog << _F("Warning: Unable to change permissions on %s: ", filename.c_str());
+             s.print_warning("Unable to change permissions on " + filename + ": ");
 	      perror ("");
 	    }
 	}
@@ -1768,11 +1796,10 @@ revoke_server_trust (
   // Make sure the given path exists.
   if (! file_exists (cert_db_path))
     {
-      if (s.verbose > 1)
-       clog << _F("Certificate database '%s' does not exist",
-                  cert_db_path.c_str()) << endl;
-      if (s.verbose)
+      if (s.verbose >= 2)
 	{
+	  clog << _F("Certificate database '%s' does not exist",
+		     cert_db_path.c_str()) << endl;
 	  for (vector<compile_server_info>::const_iterator server = server_list.begin();
 	       server != server_list.end ();
 	       ++server)
@@ -1833,7 +1860,7 @@ revoke_server_trust (
       if (! db_cert)
 	{
 	  // No trusted servers. Not an error, but issue a status message.
-	  if (s.verbose)
+	  if (s.verbose >= 2)
 	    trust_already_in_place (*server, server_list, cert_db_path, true/*revoking*/);
 	  continue;
 	}
@@ -1879,7 +1906,7 @@ revoke_server_trust (
       if (CERT_LIST_END (node, certs))
 	{
 	  // Not found. Server is already untrusted.
-	  if (s.verbose)
+	  if (s.verbose >= 2)
 	    trust_already_in_place (*server, server_list, cert_db_path, true/*revoking*/);
 	}
       else
@@ -1916,7 +1943,7 @@ get_server_info_from_db (
   // Make sure the given path exists.
   if (! file_exists (cert_db_path))
     {
-      if (s.verbose > 1)
+      if (s.verbose >= 2)
        clog << _F("Certificate database '%s' does not exist.",
                   cert_db_path.c_str()) << endl;
       return;
@@ -1938,7 +1965,7 @@ get_server_info_from_db (
   CERTCertList *certs = get_cert_list_from_db (server_cert_nickname ());
   if (! certs)
     {
-      if (s.verbose > 1)
+      if (s.verbose >= 2)
 	clog << _F("No certificate found in database %s", cert_db_path.c_str ()) << endl;
       goto cleanup;
     }
@@ -2051,6 +2078,13 @@ ostream &operator<< (ostream &s, const compile_server_info &i)
   return s;
 }
 
+ostream &operator<< (ostream &s, const vector<compile_server_info> &v)
+{
+  for (unsigned i = 0; i < v.size(); ++i)
+    s << v[i] << endl;
+  return s;
+}
+
 // Return the default server specification, used when none is given on the
 // command line.
 static string
@@ -2058,17 +2092,17 @@ default_server_spec (const systemtap_session &s)
 {
   // If the --use-server option has been used
   //   the default is 'specified'
-  // otherwise if the --unprivileged has been used
+  // otherwise if --privilege=X has been used, where X is not stapdev,
   //   the default is online,trusted,compatible,signer
   // otherwise
-  //   the default is online,compatible
+  //   the default is online,trusted,compatible
   //
   // Having said that,
   //   'online' and 'compatible' will only succeed if we have avahi
   //   'trusted' and 'signer' will only succeed if we have NSS
   //
   string working_string = "online,trusted,compatible";
-  if (s.unprivileged)
+  if (! pr_contains (s.privilege, pr_stapdev))
     working_string += ",signer";
   return working_string;
 }
@@ -2125,7 +2159,8 @@ server_spec_to_pmask (const string &server_spec)
 	}
       else
 	{
-	  clog << _F("Warning: unsupported compile server property: %s", property.c_str())
+          // XXX PR13274 needs-session to use print_warning()
+	  clog << _F("WARNING: unsupported compile server property: %s", property.c_str())
 	       << endl;
 	}
     }
@@ -2247,8 +2282,7 @@ manage_server_trust (systemtap_session &s)
       else if (*i == "no-prompt")
 	no_prompt = true;
       else
-	clog << _("Warning: Unrecognized server trust specification: ") << *i
-	     << endl;
+	s.print_warning("Unrecognized server trust specification: " + *i);
     }
   if (error)
     return;
@@ -2514,7 +2548,7 @@ get_specified_server_info (
 		  // Did we find one?
 		  if (known_servers.empty ())
 		    {
-		      if (s.verbose)
+		      if (s.verbose >= 2)
                        clog << _F("No server matching %s found", server.c_str()) << endl;
 		    }
 		  else
@@ -2595,8 +2629,8 @@ get_or_keep_trusted_server_info (
 #else // ! HAVE_NSS
       // Without NSS, we can't determine whether a server is trusted.
       // Issue a warning.
-      if (s.verbose)
-	clog << _("Unable to determine server trust as an SSL peer") << endl;
+      if (s.verbose >= 2)
+	clog << _("Unable to determine server trust as an SSL peer without NSS") << endl;
 #endif // ! HAVE_NSS
     } // Server information is not cached
 
@@ -2641,8 +2675,8 @@ get_or_keep_signing_server_info (
 #else // ! HAVE_NSS
       // Without NSS, we can't determine whether a server is a trusted
       // signer. Issue a warning.
-      if (s.verbose)
-	clog << _("Unable to determine server trust as a module signer") << endl;
+      if (s.verbose >= 2)
+	clog << _("Unable to determine server trust as a module signer without NSS") << endl;
 #endif // ! HAVE_NSS
     } // Server information is not cached
 
@@ -2708,8 +2742,8 @@ get_or_keep_compatible_server_info (
 #else // ! HAVE_AVAHI
   // Without Avahi, we can't obtain the target platform of the server.
   // Issue a warning.
-  if (s.verbose)
-    clog << _("Unable to detect server compatibility") << endl;
+  if (s.verbose >= 2)
+    clog << _("Unable to detect server compatibility without avahi") << endl;
   if (keep)
     servers.clear ();
 #endif
@@ -2814,13 +2848,6 @@ resolve_host (
       if (status == 0)
 	new_server.host_name = hbuf;
 
-      // Don't resolve to localhost or localhost.localdomain, unless that's
-      // what was asked for.
-      if ((new_server.host_name == "localhost" ||
-	   new_server.host_name == "localhost.localdomain") &&
-	  new_server.host_name != server.host_name)
-	continue;
-
       // Add the new resolved server to the list.
       add_server_info (new_server, resolved_servers);
     }
@@ -3104,8 +3131,8 @@ get_or_keep_online_server_info (
         avahi_simple_poll_free(simple_poll);
 #else // ! HAVE_AVAHI
       // Without Avahi, we can't detect online servers. Issue a warning.
-      if (s.verbose)
-	clog << _("Unable to detect online servers") << endl;
+      if (s.verbose >= 2)
+	clog << _("Unable to detect online servers without avahi") << endl;
 #endif // ! HAVE_AVAHI
     } // Server information is not cached.
 
