@@ -1,6 +1,6 @@
 /* -*- linux-c -*- 
  * Map of addresses to disallow.
- * Copyright (C) 2005-2009 Red Hat Inc.
+ * Copyright (C) 2005-20011 Red Hat Inc.
  *
  * This file is part of systemtap, and is free software.  You can
  * redistribute it and/or modify it under the terms of the GNU General
@@ -28,11 +28,7 @@ struct addr_map
   struct addr_map_entry entries[0];
 };
 
-#ifdef CONFIG_PREEMPT_RT
-static DEFINE_RAW_RWLOCK(addr_map_lock);
-#else
 static DEFINE_RWLOCK(addr_map_lock);
-#endif
 static struct addr_map* blackmap;
 
 /* Find address of entry where we can insert a new one. */
@@ -104,7 +100,8 @@ lookup_addr_aux(unsigned long addr, size_t size, struct addr_map* map)
   return 0;
 }
 
-#ifndef STP_PRIVILEGED
+#if ! STP_PRIVILEGE_CONTAINS (STP_PRIVILEGE, STP_PR_STAPDEV) && \
+    ! STP_PRIVILEGE_CONTAINS (STP_PRIVILEGE, STP_PR_STAPSYS)
 #include <asm/processor.h> /* For TASK_SIZE */
 #endif
 
@@ -117,7 +114,8 @@ lookup_bad_addr(unsigned long addr, size_t size)
   if (size == 0 || ULONG_MAX - addr < size - 1)
     return 1;
 
-#ifndef STP_PRIVILEGED
+#if ! STP_PRIVILEGE_CONTAINS (STP_PRIVILEGE, STP_PR_STAPDEV) && \
+    ! STP_PRIVILEGE_CONTAINS (STP_PRIVILEGE, STP_PR_STAPSYS)
   /* Unprivileged users must not access memory while the context
      does not refer to their own process.  */
   if (! is_myproc ())
