@@ -14,7 +14,12 @@
 #include <linux/percpu.h>
 
 static int _stp_allocated_net_memory = 0;
-#define STP_ALLOC_FLAGS (GFP_KERNEL | __GFP_NORETRY | __GFP_NOWARN)
+/* Default, and should be "safe" from anywhere. */
+#define STP_ALLOC_FLAGS ((GFP_KERNEL | __GFP_NORETRY | __GFP_NOWARN) \
+			 & ~__GFP_WAIT)
+/* May only be used in context that can sleep. __GFP_NORETRY is to
+   suppress the oom-killer from kicking in. */
+#define STP_ALLOC_SLEEP_FLAGS (GFP_KERNEL | __GFP_NORETRY)
 
 /* #define DEBUG_MEM */
 /*
@@ -35,16 +40,12 @@ static int _stp_allocated_net_memory = 0;
 static int _stp_allocated_memory = 0;
 
 #ifdef DEBUG_MEM
-#ifdef CONFIG_PREEMPT_RT
-static DEFINE_RAW_SPINLOCK(_stp_mem_lock);
-#else
 static DEFINE_SPINLOCK(_stp_mem_lock);
-#endif
 
 #define MEM_MAGIC 0xc11cf77f
 #define MEM_FENCE_SIZE 32
 
-enum _stp_memtype { MEM_KMALLOC, MEM_VMALLOC, MEM_PERCPU };
+enum _stp_memtype { MEM_KMALLOC, MEM_PERCPU };
 
 typedef struct {
 	char *alloc;
@@ -53,7 +54,6 @@ typedef struct {
 
 static const _stp_malloc_type const _stp_malloc_types[] = {
 	{"kmalloc", "kfree"},
-	{"vmalloc", "vfree"},
 	{"alloc_percpu", "free_percpu"}
 };
 
@@ -166,10 +166,6 @@ static void _stp_mem_debug_free(void *addr, enum _stp_memtype type)
 		free_percpu(addr);
 		kfree(p);
 		break;
-	case MEM_VMALLOC:
-		_stp_check_mem_fence(addr, m->len);
-		vfree(addr - MEM_FENCE_SIZE);		
-		break;
 	default:
 		printk("SYSTEMTAP ERROR: Attempted to free memory at addr %p len=%d with unknown allocation type.\n", addr, (int)m->len);
 	}
@@ -209,9 +205,6 @@ static void _stp_mem_debug_validate(void *addr)
 	case MEM_PERCPU:
 		/* do nothing */
 		break;
-	case MEM_VMALLOC:
-		_stp_check_mem_fence(addr, m->len);
-		break;
 	default:
 		printk("SYSTEMTAP ERROR: Attempted to validate memory at addr %p len=%d with unknown allocation type.\n", addr, (int)m->len);
 	}
@@ -253,7 +246,7 @@ static void _stp_mem_debug_validate(void *addr)
 #endif
 #endif
 
-static void *_stp_kmalloc(size_t size)
+static void *_stp_kmalloc_gfp(size_t size, gfp_t gfp_mask)
 {
 	void *ret;
 #ifdef STP_MAXMEMORY
@@ -263,13 +256,13 @@ static void *_stp_kmalloc(size_t size)
 	}
 #endif
 #ifdef DEBUG_MEM
-	ret = kmalloc(size + MEM_DEBUG_SIZE, STP_ALLOC_FLAGS);
+	ret = kmalloc(size + MEM_DEBUG_SIZE, gfp_mask);
 	if (likely(ret)) {
 	        _stp_allocated_memory += size;
 		ret = _stp_mem_debug_setup(ret, size, MEM_KMALLOC);
 	}
 #else
-	ret = kmalloc(size, STP_ALLOC_FLAGS);
+	ret = kmalloc(size, gfp_mask);
 	if (likely(ret)) {
 	        _stp_allocated_memory += size;
 	}
@@ -277,7 +270,12 @@ static void *_stp_kmalloc(size_t size)
 	return ret;
 }
 
-static void *_stp_kzalloc(size_t size)
+static void *_stp_kmalloc(size_t size)
+{
+	return _stp_kmalloc_gfp(size, STP_ALLOC_FLAGS);
+}
+
+static void *_stp_kzalloc_gfp(size_t size, gfp_t gfp_mask)
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
 {
 	void *ret;
@@ -288,14 +286,14 @@ static void *_stp_kzalloc(size_t size)
 	}
 #endif
 #ifdef DEBUG_MEM
-	ret = kmalloc(size + MEM_DEBUG_SIZE, STP_ALLOC_FLAGS);
+	ret = kmalloc(size + MEM_DEBUG_SIZE, gfp_mask);
 	if (likely(ret)) {
 	        _stp_allocated_memory += size;
 		ret = _stp_mem_debug_setup(ret, size, MEM_KMALLOC);
 		memset (ret, 0, size);
 	}
 #else
-	ret = kmalloc(size, STP_ALLOC_FLAGS);
+	ret = kmalloc(size, gfp_mask);
 	if (likely(ret)) {
 	        _stp_allocated_memory += size;
 		memset (ret, 0, size);
@@ -313,13 +311,13 @@ static void *_stp_kzalloc(size_t size)
 	}
 #endif
 #ifdef DEBUG_MEM
-	ret = kzalloc(size + MEM_DEBUG_SIZE, STP_ALLOC_FLAGS);
+	ret = kzalloc(size + MEM_DEBUG_SIZE, gfp_mask);
 	if (likely(ret)) {
 	        _stp_allocated_memory += size;
 		ret = _stp_mem_debug_setup(ret, size, MEM_KMALLOC);
 	}
 #else
-	ret = kzalloc(size, STP_ALLOC_FLAGS);
+	ret = kzalloc(size, gfp_mask);
 	if (likely(ret)) {
 	        _stp_allocated_memory += size;
 	}
@@ -328,28 +326,9 @@ static void *_stp_kzalloc(size_t size)
 }
 #endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15) */
 
-static void *_stp_vmalloc(unsigned long size)
+static void *_stp_kzalloc(size_t size)
 {
-	void *ret;
-#ifdef STP_MAXMEMORY
-	if ((_STP_MODULE_CORE_SIZE + _stp_allocated_memory + size)
-	    > (STP_MAXMEMORY * 1024)) {
-		return NULL;
-	}
-#endif
-#ifdef DEBUG_MEM
-	ret = __vmalloc(size + MEM_DEBUG_SIZE, STP_ALLOC_FLAGS, PAGE_KERNEL);
-	if (likely(ret)) {
-	        _stp_allocated_memory += size;
-		ret = _stp_mem_debug_setup(ret, size, MEM_VMALLOC);
-	}
-#else
-	ret = __vmalloc(size, STP_ALLOC_FLAGS, PAGE_KERNEL);
-	if (likely(ret)) {
-	        _stp_allocated_memory += size;
-	}
-#endif
-	return ret;
+  return _stp_kzalloc_gfp(size, STP_ALLOC_FLAGS);
 }
 
 #ifdef PCPU_MIN_UNIT_SIZE
@@ -358,6 +337,7 @@ static void *_stp_vmalloc(unsigned long size)
 #define _STP_MAX_PERCPU_SIZE 131072
 #endif
 
+/* Note, calls __alloc_percpu which may sleep and always uses GFP_KERNEL. */
 static void *_stp_alloc_percpu(size_t size)
 {
 	void *ret;
@@ -380,7 +360,7 @@ static void *_stp_alloc_percpu(size_t size)
 #endif
 #ifdef DEBUG_MEM
 	if (likely(ret)) {
-		struct _stp_mem_entry *m = kmalloc(sizeof(struct _stp_mem_entry), STP_ALLOC_FLAGS);
+		struct _stp_mem_entry *m = kmalloc(sizeof(struct _stp_mem_entry), GFP_KERNEL);
 		if (unlikely(m == NULL)) {
 			free_percpu(ret);
 			return NULL;
@@ -398,8 +378,9 @@ static void *_stp_alloc_percpu(size_t size)
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,12)
 #define _stp_kmalloc_node(size,node) _stp_kmalloc(size)
+#define _stp_kmalloc_node_gfp(size,node,gfp) _stp_kmalloc_gfp(size,gfp)
 #else
-static void *_stp_kmalloc_node(size_t size, int node)
+static void *_stp_kmalloc_node_gfp(size_t size, int node, gfp_t gfp_mask)
 {
 	void *ret;
 #ifdef STP_MAXMEMORY
@@ -409,19 +390,23 @@ static void *_stp_kmalloc_node(size_t size, int node)
 	}
 #endif
 #ifdef DEBUG_MEM
-	ret = kmalloc_node(size + MEM_DEBUG_SIZE, STP_ALLOC_FLAGS, node);
+	ret = kmalloc_node(size + MEM_DEBUG_SIZE, gfp_mask, node);
 	if (likely(ret)) {
 	        _stp_allocated_memory += size;
 		ret = _stp_mem_debug_setup(ret, size, MEM_KMALLOC);
 	}
 #else
-	ret = kmalloc_node(size, STP_ALLOC_FLAGS, node);
+	ret = kmalloc_node(size, gfp_mask, node);
 	if (likely(ret)) {
 	        _stp_allocated_memory += size;
 	}
 #endif
 	return ret;
 }
+static void *_stp_kmalloc_node(size_t size, int node)
+{
+	return _stp_kmalloc_node_gfp(size, node, STP_ALLOC_FLAGS);
+}
 #endif /* LINUX_VERSION_CODE */
 
 static void _stp_kfree(void *addr)
@@ -433,15 +418,6 @@ static void _stp_kfree(void *addr)
 #endif
 }
 
-static void _stp_vfree(void *addr)
-{
-#ifdef DEBUG_MEM
-	_stp_mem_debug_free(addr, MEM_VMALLOC);
-#else
-	vfree(addr);
-#endif
-}
-
 static void _stp_free_percpu(void *addr)
 {
 #ifdef DEBUG_MEM
@@ -480,10 +456,6 @@ static void _stp_mem_debug_done(void)
 			free_percpu(m->addr);
 			kfree(p);
 			break;
-		case MEM_VMALLOC:
-			_stp_check_mem_fence(m->addr, m->len);
-			vfree(m->addr - MEM_FENCE_SIZE);		
-			break;
 		default:
 			printk("SYSTEMTAP ERROR: Attempted to free memory at addr %p len=%d with unknown allocation type.\n", m->addr, (int)m->len);
 		}
