@@ -28,7 +28,7 @@ static int map_sizes[] = {
 
 static unsigned int int64_hash (const int64_t v)
 {
-	return (unsigned int)hash_long ((unsigned long)v, HASH_TABLE_BITS);
+	return (unsigned int)hash_long (((unsigned long)v) ^ stap_hash_seed, HASH_TABLE_BITS);
 }
 
 static int int64_eq_p (int64_t key1, int64_t key2)
@@ -55,14 +55,19 @@ static int str_eq_p (char *key1, char *key2)
 	return strncmp(key1, key2, MAP_STRING_LENGTH - 1) == 0;
 }
 
+static unsigned long partial_str_hash(unsigned long c, unsigned long prevhash)
+{
+	return (prevhash + (c << 4) + (c >> 4)) * 11;
+}
+
 static unsigned int str_hash(const char *key1)
 {
-	int hash = 0, count = 0;
+	unsigned long hash = 0;
+	int count = 0;
 	char *v1 = (char *)key1;
-	while (*v1 && count++ < 5) {
-		hash += *v1++;
-	}
-	return (unsigned int)hash_long((unsigned long)hash, HASH_TABLE_BITS);
+	while (*v1 && count++ < MAP_STRING_LENGTH)
+		hash = partial_str_hash(*v1++, hash);
+	return (unsigned int)hash_long(hash^stap_hash_seed, HASH_TABLE_BITS);
 }
 
 /** @addtogroup maps 
@@ -168,7 +173,16 @@ static char *_stp_key_get_str (struct map_node *mn, int n)
 static int _stp_map_init(MAP m, unsigned max_entries, int type, int key_size, int data_size, int cpu)
 {
 	int size;
-
+	size_t hash_size = sizeof(struct hlist_head) * HASH_TABLE_SIZE;
+	if(cpu < 0)
+		m->hashes = (struct hlist_head *) _stp_kmalloc_gfp(hash_size, STP_ALLOC_SLEEP_FLAGS);
+	else
+		m->hashes = (struct hlist_head *) _stp_kmalloc_node_gfp(hash_size, cpu_to_node(cpu), STP_ALLOC_SLEEP_FLAGS);
+	if(m->hashes == NULL) {
+		_stp_error("error allocating hash\n");
+		return -1;
+	}
+	memset(m->hashes, 0, hash_size);
 	m->maxnum = max_entries;
 	m->type = type;
 	if (type >= END) {
@@ -190,10 +204,11 @@ static int _stp_map_init(MAP m, unsigned max_entries, int type, int key_size, in
 		
 		
 		for (i = 0; i < max_entries; i++) {
+			/* Called at module init time, so user context.  */
 			if (cpu < 0)
-				tmp = _stp_kmalloc(size);
+				tmp = _stp_kmalloc_gfp(size, STP_ALLOC_SLEEP_FLAGS);
 			else
-				tmp = _stp_kmalloc_node(size, cpu_to_node(cpu));
+				tmp = _stp_kmalloc_node_gfp(size, cpu_to_node(cpu), STP_ALLOC_SLEEP_FLAGS);
 		
 			if (!tmp)
 				return -1;
@@ -211,7 +226,8 @@ static int _stp_map_init(MAP m, unsigned max_entries, int type, int key_size, in
 
 static MAP _stp_map_new(unsigned max_entries, int type, int key_size, int data_size)
 {
-	MAP m = (MAP) _stp_kzalloc(sizeof(struct map_root));
+	/* Called from module_init, so user context, may sleep alloc. */
+	MAP m = (MAP) _stp_kzalloc_gfp(sizeof(struct map_root), STP_ALLOC_SLEEP_FLAGS);
 	if (m == NULL)
 		return NULL;
 
@@ -229,7 +245,8 @@ static PMAP _stp_pmap_new(unsigned max_entries, int type, int key_size, int data
 	int i;
 	MAP map, m;
 
-	PMAP pmap = (PMAP) _stp_kzalloc(sizeof(struct pmap));
+	/* Called from module_init, so user context, may sleep alloc. */
+	PMAP pmap = (PMAP) _stp_kzalloc_gfp(sizeof(struct pmap), STP_ALLOC_SLEEP_FLAGS);
 	if (pmap == NULL)
 		return NULL;
 
@@ -375,6 +392,8 @@ static void __stp_map_del(MAP map)
 		list_del(p);
 		_stp_kfree(p);
 	}
+	/* free used hash */
+	_stp_kfree(map->hashes);
 }
 
 /** Deletes a map.
