@@ -254,7 +254,7 @@ int init_relayfs(void)
 		return -1;
 	}
 
-	if (send_request(STP_BULK, rqbuf, sizeof(rqbuf)) > 0)
+	if (send_request(STP_BULK, rqbuf, sizeof(rqbuf)) == 0)
 		bulkmode = 1;
 
 	for (i = 0; i < NR_CPUS; i++) {
