@@ -2,7 +2,7 @@
  *
  * staprun.c - SystemTap module loader
  *
- * Copyright (C) 2005-2010 Red Hat, Inc.
+ * Copyright (C) 2005-2011 Red Hat, Inc.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,6 +22,7 @@
 #define _XOPEN_SOURCE
 #define _BSD_SOURCE
 #include "staprun.h"
+#include "../../privilege.h"
 #include <string.h>
 #include <sys/uio.h>
 #include <glob.h>
@@ -35,7 +36,48 @@ extern long delete_module(const char *, unsigned int);
 
 int send_relocations ();
 int send_tzinfo ();
+int send_privilege_credentials ();
+int send_remote_id ();
 
+static int remove_module(const char *name, int verb);
+
+static int stap_module_inserted = -1;
+
+static void term_signal_handler(int signum __attribute ((unused)))
+{
+	if (stap_module_inserted == 0) {
+		// We have to close the control channel so that
+		// remove_module() can open it back up (which it does
+		// to make sure the module is a systemtap module).
+		close_ctl_channel();
+		remove_module(modname, 1);
+		free(modname);
+	}
+	_exit(1);
+}
+
+void setup_term_signals(void)
+{
+	sigset_t s;
+	struct sigaction a;
+
+	/* blocking all signals while we set things up */
+	sigfillset(&s);
+	sigprocmask(SIG_SETMASK, &s, NULL);
+
+	/* handle signals */
+	memset(&a, 0, sizeof(a));
+	sigfillset(&a.sa_mask);
+	a.sa_handler = term_signal_handler;
+	sigaction(SIGHUP, &a, NULL);
+	sigaction(SIGINT, &a, NULL);
+	sigaction(SIGTERM, &a, NULL);
+	sigaction(SIGQUIT, &a, NULL);
+
+	/* unblock all signals */
+	sigemptyset(&s);
+	sigprocmask(SIG_SETMASK, &s, NULL);
+}
 
 static int run_as(int exec_p, uid_t uid, gid_t gid, const char *path, char *const argv[])
 {
@@ -103,20 +145,14 @@ static int run_as(int exec_p, uid_t uid, gid_t gid, const char *path, char *cons
  */
 static int enable_uprobes(void)
 {
-	int i;
 	char *argv[10];
 	char runtimeko[2048];
-	uid_t uid = getuid();
-	gid_t gid = getgid();
+        int rc;
 
-	i = 0;
-	argv[i++] = "/bin/grep";
-	argv[i++] = "-q";
-	argv[i++] = "unregister_uprobe";
-	argv[i++] = "/proc/kallsyms";
-	argv[i] = NULL;
-	if (run_as(0, uid, gid, argv[0], argv) == 0)
-		return 0;
+        /* Formerly, we did a grep /proc/kallsyms search to see if
+           uprobes was already loaded into the kernel.  But this is
+           a race waiting to happen.  Just try to load the thing.
+           Quietly accept a -EEXIST error. */
 
         /* NB: don't use /sbin/modprobe, without more env. sanitation. */
 
@@ -124,14 +160,20 @@ static int enable_uprobes(void)
 	if (uprobes_path)
 	  snprintf (runtimeko, sizeof(runtimeko), "%s", uprobes_path);
 	else
+          /* NB: since PR5163, share/runtime/uprobes/uprobes.ko is not built 
+             by systemtap. */
 	  snprintf (runtimeko, sizeof(runtimeko), "%s/uprobes/uprobes.ko",
 		    (getenv("SYSTEMTAP_RUNTIME") ?: PKGDATADIR "/runtime"));
 	dbug(2, "Inserting uprobes module from %s.\n", runtimeko);
 	/* This module may be signed, so use insert_module to load it.  */
 	argv[0] = NULL;
-	if (insert_module(runtimeko, NULL, argv, assert_uprobes_module_permissions) == 0)
+
+	rc = insert_module(runtimeko, NULL, argv, assert_uprobes_module_permissions);
+        if ((rc == 0) || /* OK */
+            (rc == -EEXIST)) /* Someone else might have loaded it */
 		return 0;
 
+        err("Error inserting module '%s': %s\n", runtimeko, moderror(errno));
 	return 1; /* failure */
 }
 
@@ -140,14 +182,18 @@ static int insert_stap_module(void)
 	char special_options[128];
 
 	/* Add the _stp_bufsize option.  */
-	if (snprintf_chk(special_options, sizeof (special_options), "_stp_bufsize=%d", buffer_size))
+	if (snprintf_chk(special_options, sizeof (special_options),
+			 "_stp_bufsize=%d", buffer_size))
 		return -1;
 
-	return insert_module(modpath, special_options, modoptions, assert_stap_module_permissions);
+	stap_module_inserted = insert_module(modpath, special_options,
+					     modoptions,
+					     assert_stap_module_permissions);
+        if (stap_module_inserted != 0)
+                err("Error inserting module '%s': %s\n", modpath, moderror(errno));
+	return stap_module_inserted;
 }
 
-static int remove_module(const char *name, int verb);
-
 static void remove_all_modules(void)
 {
 	char *base;
@@ -213,18 +259,61 @@ static int remove_module(const char *name, int verb)
 	return 0;
 }
 
+
+/* As per PR13193, some kernels have a buggy kprobes-optimization code,
+   which results in BUG/panics in certain circumstances.  We turn off
+   kprobes optimization as a conservative measure, unless told otherwise
+   by an environment variable.
+*/
+void disable_kprobes_optimization()
+{
+        /* Test if the file exists at all. */
+        const char* proc_kprobes = "/proc/sys/debug/kprobes-optimization";
+        char prev;
+        int rc, fd;
+
+        if (getenv ("STAP_PR13193_OVERRIDE"))
+                return;
+
+        /* See the initial state; if it's already disabled, we do nothing. */
+        fd = open (proc_kprobes, O_RDONLY);
+        if (fd < 0) 
+                return;
+        rc = read (fd, &prev, sizeof(prev));
+        (void) close (fd);
+        if (rc < 1 || prev == '0') /* Already disabled or unavailable */
+                return;
+
+        fd = open (proc_kprobes, O_WRONLY);
+        if (fd < 0) 
+                return;
+        prev = '0'; /* really, next */
+        rc = write (fd, &prev, sizeof(prev));
+        (void) close (fd);
+        if (rc == 1)
+                dbug(1, "Disabled %s.\n", proc_kprobes);
+        else
+                dbug(1, "Error %d/%d disabling %s.\n", rc, errno, proc_kprobes);
+}
+
+
 int init_staprun(void)
 {
+	int rc;
 	dbug(2, "init_staprun\n");
 
 	if (mountfs() < 0)
 		return -1;
 
+	rc = 0;
 	if (delete_mod)
 		exit(remove_module(modname, 1));
 	else if (!attach_mod) {
 		if (need_uprobes && enable_uprobes() != 0)
 			return -1;
+
+                disable_kprobes_optimization();
+
 		if (insert_stap_module() < 0) {
 #ifdef HAVE_ELF_GETSHDRSTRNDX
 			if(!rename_mod && errno == EEXIST)
@@ -234,12 +323,26 @@ int init_staprun(void)
                            advise people to use -R. */
 			return -1;
 		}
-		if (send_relocations() < 0)
-			return -1;
-                if (send_tzinfo() < 0)
-                        return -1;
+		rc = init_ctl_channel (modname, 0);
+		if (rc >= 0) {
+		  /* If we are unable to send privilege credentials then we have an old
+		     (pre 1.7) stap module or a non-stap module. In either case, the privilege
+		     credentials required for loading the module have already been determined and
+		     checked (see check_groups, get_module_required_credentials).
+		  */
+		  send_privilege_credentials();
+		  rc = send_relocations();
+		  if (rc == 0) {
+		    rc = send_tzinfo();
+		    if (rc == 0 && remote_id >= 0)
+		      send_remote_id();
+		  }
+		  close_ctl_channel ();
+		}
+		if (rc != 0)
+		  remove_module(modname, 1);
 	}
-	return 0;
+	return rc;
 }
 
 int main(int argc, char **argv)
@@ -269,6 +372,7 @@ int main(int argc, char **argv)
 	}
 
 	setup_signals();
+	setup_term_signals();
 
 	parse_args(argc, argv);
 
@@ -341,26 +445,29 @@ err:
 */
 
 
-void send_a_relocation (const char* module, const char* reloc, unsigned long long address)
+int send_a_relocation (const char* module, const char* reloc, unsigned long long address)
 {
   struct _stp_msg_relocation msg;
+  int rc;
 
   if (strlen(module) >= STP_MODULE_NAME_LEN-1) {
           dbug (1, "module name too long: %s", module);
-          return; 
+          return -EINVAL; 
   }
   strncpy (msg.module, module, STP_MODULE_NAME_LEN);
   
   if (strlen(reloc) >= STP_SYMBOL_NAME_LEN-1) {
           dbug (1, "reloc name too long: %s", module);
-          return; 
+          return -EINVAL; 
   }
   strncpy (msg.reloc, reloc, STP_MODULE_NAME_LEN);
 
   msg.address = address;
 
-  send_request (STP_RELOCATION, & msg, sizeof (msg));
-  /* XXX: verify send_request RC */
+  rc = send_request (STP_RELOCATION, & msg, sizeof (msg));
+  if (rc != 0)
+    perror ("Unable to send relocation");
+  return rc;
 }
 
 
@@ -372,11 +479,16 @@ void send_a_relocation (const char* module, const char* reloc, unsigned long lon
 
 int send_relocation_kernel ()
 {
-  FILE* kallsyms = fopen ("/proc/kallsyms", "r");
+  FILE* kallsyms;
+  int rc = 0;
+
+  errno = 0;
+  kallsyms = fopen ("/proc/kallsyms", "r");
   if (kallsyms == NULL)
     {
       perror("cannot open /proc/kallsyms");
       // ... and the kernel module will almost certainly fail to initialize.
+      return errno;
     }
   else
     {
@@ -396,7 +508,9 @@ int send_relocation_kernel ()
 		  && !strcmp(line + pos, KERNEL_RELOC_SYMBOL "\n"))
                 {
                   /* NB: even on ppc, we use the _stext relocation name. */
-                  send_a_relocation ("kernel", "_stext", address);
+                  rc = send_a_relocation ("kernel", "_stext", address);
+		  if (rc != 0)
+		    break;
 
                   /* We need nothing more from the kernel. */
                   done_with_kallsyms=1;
@@ -406,20 +520,20 @@ int send_relocation_kernel ()
       free (line);
       fclose (kallsyms);
       if (!done_with_kallsyms)
-	return -1;
+	return rc;
 
       /* detect note section, send flag if there
        * NB: address=2 represents existed note, the real one in _stp_module
        */
       if (!access("/sys/kernel/notes", R_OK))
-	 send_a_relocation ("kernel", ".note.gnu.build-id", 2);
+	rc = send_a_relocation ("kernel", ".note.gnu.build-id", 2);
     }
 
-  return 0;
+  return rc;
 }
 
 
-void send_relocation_modules ()
+int send_relocation_modules ()
 {
   unsigned i = 0;
   glob_t globbuf;
@@ -427,8 +541,9 @@ void send_relocation_modules ()
   int r = glob("/sys/module/*/sections/*", GLOB_PERIOD, NULL, &globbuf);
 
   if (r == GLOB_NOSPACE || r == GLOB_ABORTED)
-    return;
+    return r;
 
+  r = 0;
   for (i=0; i<globbuf.gl_pathc; i++)
     {
       char *module_section_file;
@@ -473,7 +588,13 @@ void send_relocation_modules ()
              is open so we won't need the full name again. */
           *module_name_end = '\0';
 
-          send_a_relocation (module_name, section_name, section_address);
+          /* PR6503.  /sys/module/.../sections/...init.... sometimes contain
+             non-0 addresses, even though the respective module-initialization
+             sections were already unloaded.  We override the addresses here. */
+          if (strstr (section_name, "init.") != NULL) /* .init.text, .devinit.rodata, ... */
+             section_address = 0;
+
+          r = send_a_relocation (module_name, section_name, section_address);
         }
 
       if (strcmp (section_name, ".gnu.linkonce.this_module"))
@@ -489,9 +610,12 @@ void send_relocation_modules ()
              condition where a probe may be just starting up at the
              same time that a probeworthy module is being unloaded. */
         }
+      if (r != 0)
+	break;
     }
 
   globfree (& globbuf);
+  return r;
 }
 
 
@@ -499,25 +623,19 @@ void send_relocation_modules ()
 int send_relocations ()
 {
   int rc;
-  rc = init_ctl_channel (modname, 0);
-  if (rc < 0) goto out;
   rc = send_relocation_kernel ();
-  send_relocation_modules ();
-  close_ctl_channel ();
- out:
+  if (rc == 0)
+    rc = send_relocation_modules ();
   return rc;
 }
 
 
 int send_tzinfo ()
 {
-  int rc;
   struct _stp_msg_tzinfo tzi;
   time_t now_t;
   struct tm* now;
-
-  rc = init_ctl_channel (modname, 0);
-  if (rc < 0) goto out;
+  int rc;
 
   /* NB: This is not good enough; it sends DST-unaware numbers. */
 #if 0
@@ -531,8 +649,35 @@ int send_tzinfo ()
   tzi.tz_gmtoff = - now->tm_gmtoff;
   strncpy (tzi.tz_name, now->tm_zone, STP_TZ_NAME_LEN);
 
-  send_request(STP_TZINFO, & tzi, sizeof(tzi));
-  close_ctl_channel ();
- out:
+  rc = send_request(STP_TZINFO, & tzi, sizeof(tzi));
+  if (rc != 0)
+    perror ("Unable to send time zone information");
+  return rc;
+}
+
+int send_privilege_credentials ()
+{
+  struct _stp_msg_privilege_credentials pc;
+  int rc;
+  pc.pc_group_mask = get_privilege_credentials ();
+  rc = send_request(STP_PRIVILEGE_CREDENTIALS, & pc, sizeof(pc));
+  if (rc != 0) {
+    /* Not an error. Happens when pre 1.7 modules are loaded.  */
+    dbug (1, "Unable to send user privilege credentials");
+  }
+  return rc;
+}
+
+int send_remote_id ()
+{
+  struct _stp_msg_remote_id rem;
+  int rc;
+
+  rem.remote_id = remote_id;
+  strncpy (rem.remote_uri, remote_uri, STP_REMOTE_URI_LEN);
+  rem.remote_uri [STP_REMOTE_URI_LEN-1]='\0'; /* XXX: quietly truncate */
+  rc = send_request(STP_REMOTE_ID, & rem, sizeof(rem));
+  if (rc != 0)
+    perror ("Unable to send remote id");
   return rc;
 }
