@@ -213,6 +213,8 @@ extern const char *uprobes_path;
 extern int daemon_mode;
 extern off_t fsize_max;
 extern int fnum_max;
+extern int remote_id;
+extern const char *remote_uri;
 
 /* getopt variables */
 extern char *optarg;
