@@ -80,10 +80,10 @@ static void __stp_free_ring_buffer(void)
 static int __stp_alloc_ring_buffer(void)
 {
 	int i;
-	unsigned long buffer_size = _stp_bufsize;
+	unsigned long buffer_size = _stp_bufsize * 1024 * 1024;
 
 	if (!alloc_cpumask_var(&_stp_relay_data.trace_reader_cpumask,
-			       GFP_KERNEL))
+			       (GFP_KERNEL & ~__GFP_WAIT)))
 		goto fail;
 	cpumask_clear(_stp_relay_data.trace_reader_cpumask);
 
@@ -91,6 +91,8 @@ static int __stp_alloc_ring_buffer(void)
 		dbug_trans(1, "using default buffer size...\n");
 		buffer_size = _stp_nsubbufs * _stp_subbuf_size;
 	}
+        dbug_trans(1, "using buffer size %lu...\n", buffer_size);
+
 	/* The number passed to ring_buffer_alloc() is per cpu.  Our
 	 * 'buffer_size' is a total number of bytes to allocate.  So,
 	 * we need to divide buffer_size by the number of cpus. */
@@ -100,6 +102,21 @@ static int __stp_alloc_ring_buffer(void)
 	if (!_stp_relay_data.rb)
 		goto fail;
 
+        /* Increment _stp_allocated_memory and
+           _stp_allocated_net_memory to approximately account for
+           buffers allocated by ring_buffer_alloc. */
+        {
+#ifndef DIV_ROUND_UP
+#define DIV_ROUND_UP(n,d) (((n) + (d) - 1) / (d))
+#endif
+                u64 relay_pages;
+                relay_pages = DIV_ROUND_UP (buffer_size, PAGE_SIZE);
+                if (relay_pages < 2) relay_pages = 2;
+                relay_pages *= num_online_cpus();
+                _stp_allocated_net_memory += relay_pages * PAGE_SIZE;
+                _stp_allocated_memory += relay_pages * PAGE_SIZE;
+        }
+
 	dbug_trans(0, "size = %lu\n", ring_buffer_size(_stp_relay_data.rb));
 	return 0;
 
@@ -252,7 +269,11 @@ static void _stp_ring_buffer_consume(struct _stp_iterator *iter)
 {
 	_stp_ring_buffer_iterator_increment(iter);
 	_stp_ring_buffer_disable_cpu();
+#ifdef STAPCONF_RING_BUFFER_LOST_EVENTS
+	ring_buffer_consume(_stp_relay_data.rb, iter->cpu, &iter->ts, NULL);
+#else
 	ring_buffer_consume(_stp_relay_data.rb, iter->cpu, &iter->ts);
+#endif
 	_stp_ring_buffer_enable_cpu();
 	atomic_dec(&iter->nr_events);
 }
@@ -288,7 +309,11 @@ _stp_peek_next_event(struct _stp_iterator *iter, int cpu, u64 *ts)
 	if (iter->buffer_iter[cpu])
 		event = ring_buffer_iter_peek(iter->buffer_iter[cpu], ts);
 	else
+#ifdef STAPCONF_RING_BUFFER_LOST_EVENTS
+		event = ring_buffer_peek(_stp_relay_data.rb, cpu, ts, NULL);
+#else
 		event = ring_buffer_peek(_stp_relay_data.rb, cpu, ts);
+#endif
 	_stp_ring_buffer_enable_cpu();
 	return event;
 }
@@ -383,11 +408,19 @@ _stp_buffer_iter_start(struct _stp_iterator *iter)
 	int cpu_file = iter->cpu_file;
 
 	iter->buffer_iter[cpu_file]
+#ifdef STAPCONF_RING_BUFFER_READ_PREPARE
+	    = ring_buffer_read_prepare(_stp_relay_data.rb, cpu_file);
+#else
 	    = ring_buffer_read_start(_stp_relay_data.rb, cpu_file);
+#endif
 	if (iter->buffer_iter[cpu_file] == NULL) {
 		dbug_trans(0, "buffer_iter[%d] was NULL\n", cpu_file);
 		return 1;
 	}
+#ifdef STAPCONF_RING_BUFFER_READ_PREPARE
+	ring_buffer_read_prepare_sync();
+	ring_buffer_read_start(iter->buffer_iter[cpu_file]);
+#endif
 	dbug_trans(0, "iterator(s) started\n");
 	return 0;
 #else
@@ -395,13 +428,23 @@ _stp_buffer_iter_start(struct _stp_iterator *iter)
 
 	for_each_online_cpu(cpu) {
 		iter->buffer_iter[cpu]
+#ifdef STAPCONF_RING_BUFFER_READ_PREPARE
+		    = ring_buffer_read_prepare(_stp_relay_data.rb, cpu);
+#else
 		    = ring_buffer_read_start(_stp_relay_data.rb, cpu);
+#endif
 		if (iter->buffer_iter[cpu] == NULL) {
 			dbug_trans(0, "buffer_iter[%d] was NULL\n", cpu);
 			_stp_buffer_iter_finish(iter);
 			return 1;
 		}
 	}
+#ifdef STAPCONF_RING_BUFFER_READ_PREPARE
+	ring_buffer_read_prepare_sync();
+	for_each_online_cpu(cpu) {
+		ring_buffer_read_start(iter->buffer_iter[cpu]);
+	}
+#endif
 	dbug_trans(0, "iterator(s) started\n");
 	return 0;
 #endif
