@@ -13,12 +13,22 @@
 #ifndef _STP_UNWIND_H_
 #define _STP_UNWIND_H_
 
-#ifdef STP_USE_DWARF_UNWINDER
+struct unwind_frame_info
+{
+    struct pt_regs regs;
+    unsigned call_frame:1;
+};
 
 #if defined (__x86_64__)
 #include "x86_64.h"
 #elif  defined (__i386__)
 #include "i386.h"
+#elif defined (__powerpc64__)
+#include "ppc64.h"
+#elif defined (__s390x__)
+#include "s390x.h"
+#elif defined (__arm__)
+#include "arm.h"
 #else
 #error "Unsupported dwarf unwind architecture"
 #endif
@@ -30,6 +40,10 @@
 #define BUILD_BUG_ON_ZERO(e) (sizeof(char[1 - 2 * !!(e)]) - 1)
 #endif
 
+#ifndef FIELD_SIZEOF
+#define FIELD_SIZEOF(t, f) (sizeof(((t*)0)->f))
+#endif
+
 
 #define EXTRA_INFO(f) { \
 		BUILD_BUG_ON_ZERO(offsetof(struct unwind_frame_info, f) \
@@ -50,10 +64,39 @@ static const struct {
 #undef PTREGS_INFO
 #undef EXTRA_INFO
 
+/* The reg_info array assumes dwarf register numbers start at zero and
+   are consecutive.  If that isn't the case for some architecture (e.g. ppc)
+   then redefine to map the given dwarf register number to the actual
+   reg_info index.  */
+#ifndef DWARF_REG_MAP
+#define DWARF_REG_MAP(r) r
+#endif
+
+/* The number of real registers in the register map. These are all assumed
+   to be the Same in the new frame. All others will be Unknown untill they
+   have been explictly set. (e.g. the x86 return register). */
+#ifndef UNW_NR_REAL_REGS
+#define UNW_NR_REAL_REGS ARRAY_SIZE(reg_info)
+#endif
+
 #ifndef REG_INVALID
 #define REG_INVALID(r) (reg_info[r].width == 0)
 #endif
 
+/* Whether the stack pointer should be set from the CFA.
+   If this isn't what the architecture wants, then it should define
+   this as zero.  */
+#ifndef UNW_SP_FROM_CFA
+#define UNW_SP_FROM_CFA 1
+#endif
+
+/* Whether the instruction pointer should be set from the return address
+   column.  If this isn't what the architecture wants, then it should
+   define this as zero.  */
+#ifndef UNW_PC_FROM_RA
+#define UNW_PC_FROM_RA 1
+#endif
+
 #define DW_CFA_nop                          0x00
 #define DW_CFA_set_loc                      0x01
 #define DW_CFA_advance_loc1                 0x02
@@ -250,6 +293,7 @@ typedef   signed long sleb128_t;
 
 struct unwind_item {
 	enum item_location {
+		Same,
 		Nowhere,
 		Memory,
 		Register,
@@ -280,9 +324,14 @@ struct unwind_state {
 	sleb128_t dataAlign;
 	unsigned stackDepth:8;
 	struct unwind_reg_state reg[STP_MAX_STACK_DEPTH];
+	struct unwind_item cie_regs[ARRAY_SIZE(reg_info)];
+};
+
+struct unwind_context {
+    struct unwind_frame_info info;
+    struct unwind_state state;
 };
 
 static const struct cfa badCFA = { ARRAY_SIZE(reg_info), 1 };
 
-#endif /* STP_USE_DWARF_UNWINDER */
 #endif /*_STP_UNWIND_H_*/
