@@ -22,15 +22,22 @@
 static void _stp_vma_match_vdso(struct task_struct *tsk)
 {
 /* vdso is arch specific */
-#ifdef STAPCONF_MM_CONTEXT_VDSO
+#if defined(STAPCONF_MM_CONTEXT_VDSO) || defined(STAPCONF_MM_CONTEXT_VDSO_BASE)
   int i, j;
   if (tsk->mm)
     {
       struct _stp_module *found = NULL;
+
+#ifdef STAPCONF_MM_CONTEXT_VDSO
       unsigned long vdso_addr = (unsigned long) tsk->mm->context.vdso;
+#else
+      unsigned long vdso_addr = tsk->mm->context.vdso_base;
+#endif
+
 #ifdef DEBUG_TASK_FINDER_VMA
       _dbug("tsk: %d vdso: 0x%lx\n", tsk->pid, vdso_addr);
 #endif
+
       for (i = 0; i < _stp_num_modules && found == NULL; i++) {
 	struct _stp_module *m = _stp_modules[i];
 	if (m->path[0] == '/'
@@ -39,9 +46,10 @@ static void _stp_vma_match_vdso(struct task_struct *tsk)
 	  {
 	    unsigned long notes_addr;
 	    int all_ok = 1;
-	    notes_addr = vdso_addr + m->build_id_offset - m->build_id_len;
+	    notes_addr = vdso_addr + m->build_id_offset;
 #ifdef DEBUG_TASK_FINDER_VMA
-	    _dbug("notes_addr %s: 0x%lx\n", m->name, notes_addr);
+	    _dbug("notes_addr %s: 0x%lx + 0x%lx = 0x%lx (len: %x)\n", m->name,
+		  vdso_addr, m->build_id_offset, notes_addr, m->build_id_len);
 #endif
 	    for (j = 0; j < m->build_id_len; j++)
 	      {
@@ -53,7 +61,7 @@ static void _stp_vma_match_vdso(struct task_struct *tsk)
 		if (rc || b != m->build_id_bits[j])
 		  {
 #ifdef DEBUG_TASK_FINDER_VMA
-		    _dbug("darn, not equal (rc=%d) at %d (%d != %d)\n",
+		    _dbug("darn, not equal (rc=%d) at %d (0x%x != 0x%x)\n",
 			  rc, j, b, m->build_id_bits[j]);
 #endif
 		    all_ok = 0;
@@ -124,27 +132,39 @@ static int _stp_vma_mmap_cb(struct stap_task_finder_target *tgt,
 	// We are only interested in the first load of the whole module that
 	// is executable. We register whether or not we know the module,
 	// so we can later lookup the name given an address for this task.
-	if (path != NULL && offset == 0 && (vm_flags & VM_EXEC)) {
+	if (path != NULL && offset == 0 && (vm_flags & VM_EXEC)
+	    && stap_find_vma_map_info(tsk, addr, NULL, NULL, NULL, NULL) != 0) {
 		for (i = 0; i < _stp_num_modules; i++) {
 			if (strcmp(path, _stp_modules[i]->path) == 0)
 			{
+			  unsigned long vm_start = 0;
+			  unsigned long vm_end = 0;
 #ifdef DEBUG_TASK_FINDER_VMA
 			  _stp_dbug(__FUNCTION__, __LINE__,
 				    "vm_cb: matched path %s to module (sec: %s)\n",
 				    path, _stp_modules[i]->sections[0].name);
 #endif
 			  module = _stp_modules[i];
-			  /* XXX We really only need to register .dynamic
-			     sections, but .absolute exes are also necessary
-			     atm. */
-			  res = stap_add_vma_map_info(tsk->group_leader,
-						      addr, addr + length,
-						      path, module);
-			  /* Warn, but don't error out. */
+			  /* Make sure we really don't know about this module
+			     yet.  If we do know, we might want to extend
+			     the coverage. */
+			  res = stap_find_vma_map_info_user(tsk->group_leader,
+							    module,
+							    &vm_start, &vm_end,
+							    NULL);
+			  if (res == -ESRCH)
+			    res = stap_add_vma_map_info(tsk->group_leader,
+						        addr, addr + length,
+						        path, module);
+			  else if (res == 0 && vm_end + 1 == addr)
+			    res = stap_extend_vma_map_info(tsk->group_leader,
+							   vm_start,
+							   addr + length);
+			  /* VMA entries are allocated dynamically, this is fine,
+			   * since we are in a task_finder callback, which is in
+			   * user context. */
 			  if (res != 0) {
-				_stp_warn ("Couldn't register module '%s' for pid %d (%d)\n", _stp_modules[i]->path, tsk->group_leader->pid, res);
-				if (res == -ENOMEM)
-					_stp_warn ("Try increasing -DTASK_FINDER_VMA_ENTRY_ITEMS\n");
+				_stp_error ("Couldn't register module '%s' for pid %d (%d)\n", _stp_modules[i]->path, tsk->group_leader->pid, res);
 			  }
 			  return 0;
 			}
@@ -162,12 +182,26 @@ static int _stp_vma_mmap_cb(struct stap_task_finder_target *tgt,
 						addr + length, path, NULL);
 #ifdef DEBUG_TASK_FINDER_VMA
 		    _stp_dbug(__FUNCTION__, __LINE__,
-			      "registered '%s' for %d (res:%d)\n",
+			      "registered '%s' for %d (res:%d) [%lx-%lx]\n",
 			      path, tsk->group_leader->pid,
-			      res);
+			      res, addr, addr + length);
 #endif
 		  }
 
+	} else if (path != NULL) {
+		// Once registered, we may want to extend an earlier
+		// registered region. A segment might be mapped with
+		// different flags for different offsets. If so we want
+		// to record the extended range so we can address more
+		// precisely to module names and symbols.
+		res = stap_extend_vma_map_info(tsk->group_leader,
+					       addr, addr + length);
+#ifdef DEBUG_TASK_FINDER_VMA
+		_stp_dbug(__FUNCTION__, __LINE__,
+			  "extended '%s' for %d (res:%d) [%lx-%lx]\n",
+			  path, tsk->group_leader->pid,
+			  res, addr, addr + length);
+#endif
 	}
 	return 0;
 }
@@ -187,7 +221,7 @@ static int _stp_vma_munmap_cb(struct stap_task_finder_target *tgt,
 static int _stp_vma_init(void)
 {
         int rc = 0;
-#if defined(CONFIG_UTRACE)
+#ifdef HAVE_TASK_FINDER
         static struct stap_task_finder_target vmcb = {
                 // NB: no .pid, no .procname filters here.
                 // This means that we get a system-wide mmap monitoring
@@ -206,7 +240,11 @@ static int _stp_vma_init(void)
                 .munmap_callback = &_stp_vma_munmap_cb,
                 .mprotect_callback = NULL
         };
-	stap_initialize_vma_map ();
+	rc = stap_initialize_vma_map ();
+	if (rc != 0) {
+		_stp_error("Couldn't initialize vma map: %d\n", rc);
+		return rc;
+	}
 #ifdef DEBUG_TASK_FINDER_VMA
 	_stp_dbug(__FUNCTION__, __LINE__,
 		  "registering vmcb (_stap_target: %d)\n", _stp_target);
@@ -218,4 +256,12 @@ static int _stp_vma_init(void)
 	return rc;
 }
 
+/* Get rid of the vma tracker (memory). */
+static void _stp_vma_done(void)
+{
+#if defined(CONFIG_UTRACE)
+	stap_destroy_vma_map();
+#endif
+}
+
 #endif /* _STP_VMA_C_ */
