@@ -248,12 +248,14 @@ procfs_derived_probe_group::emit_module_decls (systemtap_session& s)
     {
       s.op->newline() << "struct _stp_procfs_data pdata;";
 
-      common_probe_entryfn_prologue (s.op, "STAP_SESSION_RUNNING", "spp->read_probe");
+      common_probe_entryfn_prologue (s.op, "STAP_SESSION_RUNNING",
+				     "spp->read_probe",
+				     "_STP_PROBE_HANDLER_PROCFS");
 
       s.op->newline() << "pdata.buffer = spp->buffer;";
       s.op->newline() << "pdata.bufsize = spp->bufsize;";
-      s.op->newline() << "if (c->data == NULL)";
-      s.op->newline(1) << "c->data = &pdata;";
+      s.op->newline() << "if (c->ips.procfs_data == NULL)";
+      s.op->newline(1) << "c->ips.procfs_data = &pdata;";
       s.op->newline(-1) << "else {";
 
       s.op->newline(1) << "if (unlikely (atomic_inc_return (& skipped_count) > MAXSKIPPED)) {";
@@ -268,11 +270,11 @@ procfs_derived_probe_group::emit_module_decls (systemtap_session& s)
       s.op->newline() << "(*spp->read_probe->ph) (c);";
 
       // Note that _procfs_value_set copied string data into spp->buffer
-      s.op->newline() << "c->data = NULL;";
+      s.op->newline() << "c->ips.procfs_data = NULL;";
       s.op->newline() << "spp->needs_fill = 0;";
       s.op->newline() << "spp->count = strlen(spp->buffer);";
 
-      common_probe_entryfn_epilogue (s.op);
+      common_probe_entryfn_epilogue (s.op, true, s.suppress_handler_errors);
 
       s.op->newline() << "if (spp->needs_fill) {";
       s.op->newline(1) << "spp->needs_fill = 0;";
@@ -292,7 +294,9 @@ procfs_derived_probe_group::emit_module_decls (systemtap_session& s)
     {
       s.op->newline() << "struct _stp_procfs_data pdata;";
 
-      common_probe_entryfn_prologue (s.op, "STAP_SESSION_RUNNING", "spp->write_probe");
+      common_probe_entryfn_prologue (s.op, "STAP_SESSION_RUNNING",
+				     "spp->write_probe",
+				     "_STP_PROBE_HANDLER_PROCFS");
 
       // We've got 2 problems here.  The data count could be greater
       // than MAXSTRINGLEN or greater than the bufsize (if the same
@@ -308,8 +312,8 @@ procfs_derived_probe_group::emit_module_decls (systemtap_session& s)
       s.op->newline() << "pdata.buffer = (char *)buf;";
       s.op->newline() << "pdata.count = count;";
 
-      s.op->newline() << "if (c->data == NULL)";
-      s.op->newline(1) << "c->data = &pdata;";
+      s.op->newline() << "if (c->ips.procfs_data == NULL)";
+      s.op->newline(1) << "c->ips.procfs_data = &pdata;";
       s.op->newline(-1) << "else {";
 
       s.op->newline(1) << "if (unlikely (atomic_inc_return (& skipped_count) > MAXSKIPPED)) {";
@@ -323,12 +327,12 @@ procfs_derived_probe_group::emit_module_decls (systemtap_session& s)
       // call probe function
       s.op->newline() << "(*spp->write_probe->ph) (c);";
 
-      s.op->newline() << "c->data = NULL;";
+      s.op->newline() << "c->ips.procfs_data = NULL;";
       s.op->newline() << "if (c->last_error == 0) {";
       s.op->newline(1) << "retval = count;";
       s.op->newline(-1) << "}";
 
-      common_probe_entryfn_epilogue (s.op);
+      common_probe_entryfn_epilogue (s.op, true, s.suppress_handler_errors);
     }
 
   s.op->newline() << "return retval;";
@@ -428,7 +432,7 @@ procfs_var_expanding_visitor::visit_target_symbol (target_symbol* e)
       ec->tok = e->tok;
 
       string fname;
-      string locvalue = "CONTEXT->data";
+      string locvalue = "CONTEXT->ips.procfs_data";
 
       if (! lvalue)
         {
