@@ -55,11 +55,24 @@ static void call10(int a, int b, int c, int d, int e, int f, int g, int h, int i
   STAP_PROBE10(test, mark_j, a, b, c, d, e, f, g, h, i, j);
 }
 
+/* I CAN HAZ MOAR ARGS? */
+#ifdef STAP_PROBE12
+static void call11(int a, int b, int c, int d, int e, int f, int g, int h, int i, int j, int k)
+{
+  STAP_PROBE11(test, mark_k, a, b, c, d, e, f, g, h, i, j, k);
+}
+
+static void call12(int a, int b, int c, int d, int e, int f, int g, int h, int i, int j, int k, int l)
+{
+  STAP_PROBE12(test, mark_l, a, b, c, d, e, f, g, h, i, j, k, l);
+}
+#endif
+
 int
 main (int argc, char **argv)
 {
-  int a, b, c, d, e, f, g, h, i, j;
-  a = 1; b = 2; c = 3; d = 4; e = 5; f = 6; g = 7; h = 8; i = 9; j = 10;
+  int a, b, c, d, e, f, g, h, i, j, k, l;
+  a = 1; b = 2; c = 3; d = 4; e = 5; f = 6; g = 7; h = 8; i = 9; j = 10; k = 11; l = 12;
   call0();
   call1(a);
   call2(a, b);
@@ -71,6 +84,12 @@ main (int argc, char **argv)
   call8(a, b, c, d, e, f, g, h);
   call9(a, b, c, d, e, f, g, h, i);
   call10(a, b, c, d, e, f, g, h, i, j);
+#ifdef STAP_PROBE12
+  call11(a, b, c, d, e, f, g, h, i, j, k);
+  call12(a, b, c, d, e, f, g, h, i, j, k, l);
+#else
+  (void)(k + l);
+#endif
   (void) argv;
   (void) argc;
   return 0;
