@@ -27,6 +27,8 @@ class translator_output
   unsigned tablevel;
 
 public:
+  std::string filename;
+
   translator_output (std::ostream& file);
   translator_output (const std::string& filename, size_t bufsize = 8192);
   ~translator_output ();
@@ -80,8 +82,9 @@ struct unparser
   // static void function_NAME (context* c);
 
   virtual void emit_module_init () = 0;
+  virtual void emit_module_refresh () = 0;
   virtual void emit_module_exit () = 0;
-  // XXX
+  // startup, probe refresh/activation, shutdown
 
   virtual void emit_function (functiondecl* v) = 0;
   // void function_NAME (struct context* c) {
